/* graph.cc
 * Copyright (C) 2001-2 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gconf/gconf-client.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <pwd.h>
#include "header.h"
#include "graph.h"
#include "misc.h"

char *oid_values[] = {
".iso.3.6.1.2.1.2.2.1.8",				// ifOperStatus
".iso.3.6.1.2.1.2.2.1.10",			// ifInOctets
".iso.3.6.1.2.1.2.2.1.11",			// ifInUcastPkts
".iso.3.6.1.2.1.2.2.1.14",			// ifInErrors
".iso.3.6.1.2.1.2.2.1.16",			// ifOutOctets
".iso.3.6.1.2.1.2.2.1.17",			// ifOutUcastPkts
".iso.3.6.1.2.1.2.2.1.20",			// ifOutErrors
".iso.3.6.1.2.1.31.1.1.1.2",		// ifInMulticastPkts
".iso.3.6.1.2.1.31.1.1.1.3",		// ifInBroadcastPkts
".iso.3.6.1.2.1.31.1.1.1.4",		// ifOutMulticastPkts
".iso.3.6.1.2.1.31.1.1.1.5",		// ifOutBroadcastPkts
".iso.3.6.1.2.1.16.1.1.1.2",		// etherStatsDataSource
".iso.3.6.1.2.1.16.1.1.1.3",		// etherStatsDropEvents
".iso.3.6.1.2.1.16.1.1.1.8",		// etherStatsCRCAlignErrors
".iso.3.6.1.2.1.16.1.1.1.9",		// etherStatsUndersizePkts
".iso.3.6.1.2.1.16.1.1.1.10",		// etherStatsOversizePkts
".iso.3.6.1.2.1.16.1.1.1.11",		// etherStatsFragments
".iso.3.6.1.2.1.16.1.1.1.12",		// etherStatsJabbers
".iso.3.6.1.2.1.16.1.1.1.13"		// etherStatsCollisions
};

int oid_types[] = { 4, 1, 1, 3, 1, 1, 3, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3 };

unsigned long get_value(unsigned long oval, unsigned long nval, int time_diff, int rate)
{
	unsigned long tmp;

	// FIXME a hack for testing generated data
//	rate = 1;

	if (oval > nval)
		tmp = 0xffffffff - (oval - nval) + 1;
	else
		tmp = nval - oval;
	if (time_diff != 0)
	{
		if (rate==1)	// rate of change per second
			tmp = tmp/(unsigned long)time_diff;
		else
		if (rate==2)	// rate of change over five minutes
			tmp = (tmp*300)/(unsigned long)time_diff;
		else
		if (rate==4)	// return the new value
			tmp = nval;
	}else
		tmp = 0;
	// rate==3 just return tmp
	return tmp;
}

void Dvice::clear()
{
	Link *ln;
	Iface *iface;

	grabtime.empty();
	for (ln=interfaces.get_head(); ln!=NULL; ln=ln->next)
	{ iface = (Iface*) ln->item; delete iface; }
	interfaces.empty();
}

Graph::Graph()
{
	packets.record_size = GRAPH_PACKET_RECORD_SIZE;
	errors.record_size = GRAPH_ERROR_RECORD_SIZE;
}

Graph::~Graph()
{
}

// we load the current data and the old data files
// store in two device objects
// if an entry is missing from either leave data value in output as 0
// if status of either is down don't record data
// if both entries the output value is the difference (fix rounding errors)
// move current data overwriting old data

void Graph::main(char *datadir, char *device)
{
	char buf[256];
	int a, b;

	// grab_gconf already called to initialise global values from gconf

	strcpy(name, device);
	sprintf(buf, "%s/raw/%s", datadir, device);
	a = load(&d1, buf);
	sprintf(buf, "%s/raw/%s.old", datadir, device);
	b = load(&d2, buf);

printf("a=%d b=%d\n", a, b);
	if ((a==1) && (b==1))
	{
		printf("GOT BOTH FILES %s\n", device);
		print();

		if (strcmp(device, "x.x.x.x")==0)
		{
			printf("New Time = %s\n", d1.grabtime.date_to_str(1));
			printf("Old Time = %s\n", d2.grabtime.date_to_str(1));
		}

		save(datadir, device);
	}
}

int Graph::load(Dvice *dv, char *fname)
{
	FILE *f;
	char buf[256], var[256], val[256];
	int i;

//	printf("Loading %s\n", fname);
	f = fopen(fname, "r");
	if (!f) return 0;

	dv->clear();

	fscanf(f, "%[^\n]\n", buf);
	dv->grabtime.setstring(buf);
	while(fscanf(f, "%[^\n]\n", buf)==1)
	{
		if (sscanf(buf, "%[^ ] %[^\n]", var, val)==2)
		{
			// check for status, packet and error values
			for (i=0; i<GRAPH_NUM_TOTAL; i++)
			{
				if (starts_with(oid_values[i], var))
				{
					add_to_interface(var, val, dv, i);
				}
			}
		}
	}

	if (dv->grabtime.zero()) return 0;

	fclose(f);

	return 1;
}

void Graph::add_to_interface(char *var, char *val, Dvice *dv, int pos)
{
	char vartype[256];
	unsigned long s;
	int ifnum;
	Iface *iface, *iffound;
	Link *ln;

// printf("ADD TO INTERFACE\n%s\n%s\n", var, gio.global_types[pos].oid);

	// start by extracting the interface number
	strcpy(vartype, var+strlen(oid_values[pos])+1);
	if (sscanf(vartype, "%d", &ifnum)!=1) return;

	// find an interface with same ifnum
	iffound = NULL;
	for (ln=dv->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		iface = (Iface*) ln->item;
		if (iface->index == ifnum) iffound = iface;
	}
	if (iffound == NULL)
	{
		iface = new Iface();
		iface->index = ifnum;
		dv->interfaces.tag(iface);
	}else
		iface = iffound;

	if (sscanf(val, "%lu", &s)==1)
		iface->values[pos] = s;
}

void Graph::print()
{
	int i, j;
	Iface *if1, *if2;
	TimeStamp time_diff;

/*
	printf("\n\nDevice %s\n", name);
	for (i=0; i<d1.interfaces.size(); i++)
	{
		if1 = (Iface*) d1.interfaces.get_item(i);
		if2 = (Iface*) d2.interfaces.get_item(i);
		time_diff.ts = d1.grabtime.compare(d2.grabtime);
		printf("Index %8d Timediff %lu\n", if1->index, time_diff.ts);
		for (j=0; j<GRAPH_NUM_TOTAL; j++)
		{
			printf("%20s Old %12lu, ", oid_values[j]+20, if2->values[j]);
			printf("New %12lu Value %12lu\n", if1->values[j],
				get_value(if2->values[j], if1->values[j], time_diff.ts, oid_types[j]));
		}
	}
*/
}

// we need to save two files, one with general values that will be kept
// for a long time, the other with errors which are kept a shorter time
// (weeks not years)
void Graph::save(char *datadir, char *device)
{
	struct stat statbuf;
	Link *ln, *ln2;
	Iface *i1, *i2;
	char buf[256], buf2[256];
	int i, found;
	int gstart;
	int estart;
	TimeStamp time_diff;
	gstart = 1;		// status is first slot
	estart = 1+GRAPH_NUM_PACKETS;
	time_diff.ts = d1.grabtime.compare(d2.grabtime);
	if (time_diff.ts==0) return;
	for (ln=d1.interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		i1 = (Iface*) ln->item;
		found = 0;
		packets.current = d1.grabtime;
		packets.status = (i1->values[0] == 1)?'0':'1';
		errors.current = d1.grabtime;
		errors.status = (i1->values[0] == 1)?'0':'1';
		for (ln2=d2.interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			i2 = (Iface*) ln2->item;
			if (i1->index == i2->index)
			{
				for (i=gstart; i<gstart+GRAPH_NUM_PACKETS; i++)
					packets.value[i-gstart] = get_value(i2->values[i], i1->values[i],
						time_diff.ts, oid_types[i]);
				for (i=estart; i<estart+GRAPH_NUM_ERRORS; i++)
					errors.value[i-estart] = get_value(i2->values[i], i1->values[i],
						time_diff.ts, oid_types[i]);
				found = 1;
//				printf ("CHILD: Found pair of matching interfaces\n");
			}
		}
		if (found)
		{
			sprintf(buf, "%s/graph", datadir);
			if (stat(buf, &statbuf) == -1) mkdir(buf, 493); // rwxr xr x
			sprintf(buf, "%s/graph/%s", datadir, device);
			if (stat(buf, &statbuf) == -1) mkdir(buf, 493); // rwxr xr x
			sprintf(buf, "%s/graph/%s/v%03d", datadir, device, i1->index);
//			printf ("CHILD: Going int graphio class to do add\n");
			packets.add(buf);
			sprintf(buf, "%s/graph/%s/e%03d", datadir, device, i1->index);
			errors.add(buf);
		}
	}
	sprintf(buf, "%s/graph/%s/v001", datadir, device);
	packets.print(buf);
}

void Graph::grab_gconf()
{
	GConfClient *client;
	GSList *slist, *tmp;
	gint num, i;
	gchar *str;
	gboolean tmpbool;

	client = gconf_client_get_default();

	slist = gconf_client_get_list(client, CONFIG_SAMPLE_FREQUENCY,
		GCONF_VALUE_INT, NULL);
	tmp = slist;
	i=0;
	while (tmp!=NULL)
	{
		num = GPOINTER_TO_INT(tmp->data);
		packets.global_blocks[i].gap = num;
		i++;
		tmp = g_slist_next(tmp);
	}
	g_slist_free(slist);
	slist = gconf_client_get_list(client, CONFIG_SAMPLE_DURATION,
		GCONF_VALUE_INT, NULL);
	tmp = slist;
	i=0;
	while (tmp!=NULL)
	{
		num = GPOINTER_TO_INT(tmp->data);
		packets.global_blocks[i].howlong = num;
		i++;
		tmp = g_slist_next(tmp);
	}
	g_slist_free(slist);
	packets.num_global_blocks = i;
	packets.num_values = GRAPH_NUM_PACKETS;

	str = gconf_client_get_string(client, CONFIG_SAMPLE_ERRORS, NULL);
	errors.global_blocks[0].howlong = timestr_to_int(str);
	str = gconf_client_get_string(client, CONFIG_SAMPLE_RATE, NULL);
	errors.global_blocks[0].gap = timestr_to_int(str);
	errors.num_global_blocks = 1;
	errors.num_values = GRAPH_NUM_ERRORS;

	tmpbool = gconf_client_get_bool(client, CONFIG_SAMPLE_REPLACE, NULL);
	packets.replace_flag = tmpbool;
	errors.replace_flag = tmpbool;

//	printf("GRABBED GCONF VALUES\n");
}

